# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 NeonGecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from abc import abstractmethod, ABC
from threading import Thread, Lock
from time import time
from typing import Optional

from neon_mq_connector.connector import MQConnector
from neon_mq_connector.utils.rabbit_utils import create_mq_callback
from neon_utils.logger import LOG

from neon_data_models.models.api.mq import (
    LLMProposeResponse,
    LLMDiscussResponse,
    LLMVoteResponse,
)

from neon_llm_core.utils.config import load_config
from neon_llm_core.llm import NeonLLM
from neon_llm_core.utils.constants import LLM_VHOST
from neon_llm_core.utils.personas.provider import PersonasProvider


class NeonLLMMQConnector(MQConnector, ABC):
    """
        Module for processing MQ requests to Fast Chat LLM
    """

    async_consumers_enabled = True

    def __init__(self, config: Optional[dict] = None):
        self.service_name = f'neon_llm_{self.name}'

        self.ovos_config = config or load_config()
        mq_config = self.ovos_config.get("MQ", dict())
        super().__init__(config=mq_config, service_name=self.service_name)
        self.vhost = LLM_VHOST

        self.register_consumers()
        self._model = None
        self._bots = list()
        self._persona_update_lock = Lock()
        self._last_persona_update = time()
        self._personas_provider = PersonasProvider(service_name=self.name,
                                                   ovos_config=self.ovos_config)

    def register_consumers(self):
        for idx in range(self.model_config.get("num_parallel_processes", 1)):
            self.register_consumer(name=f"neon_llm_{self.name}_ask_{idx}",
                                   vhost=self.vhost,
                                   queue=self.queue_ask,
                                   callback=self.handle_request,
                                   on_error=self.default_error_handler,)
        self.register_consumer(name=f'neon_llm_{self.name}_score',
                               vhost=self.vhost,
                               queue=self.queue_score,
                               callback=self.handle_score_request,
                               on_error=self.default_error_handler,)
        self.register_consumer(name=f'neon_llm_{self.name}_discussion',
                               vhost=self.vhost,
                               queue=self.queue_opinion,
                               callback=self.handle_opinion_request,
                               on_error=self.default_error_handler,)
        self.register_subscriber(name=f'neon_llm_{self.name}_update_persona',
                                 vhost=self.vhost,
                                 exchange=self.exchange_persona_updated,
                                 callback=self.handle_persona_update,
                                 on_error=self.default_error_handler)
        self.register_subscriber(name=f'neon_llm_{self.name}_delete_persona',
                                 vhost=self.vhost,
                                 exchange=self.exchange_persona_deleted,
                                 callback=self.handle_persona_delete,
                                 on_error=self.default_error_handler)

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    def model_config(self):
        if f"LLM_{self.name.upper()}" not in self.ovos_config:
            LOG.warning(f"No config for {self.name} found in "
                        f"{list(self.ovos_config.keys())}")
        return self.ovos_config.get(f"LLM_{self.name.upper()}", dict())
    
    @property
    def queue_ask(self):
        return f"{self.name}_input"
    
    @property
    def queue_score(self):
        return f"{self.name}_score_input"
    
    @property
    def queue_opinion(self):
        return f"{self.name}_discussion_input"

    @property
    def exchange_persona_updated(self):
        return f"{self.name}_persona_updated"

    @property
    def exchange_persona_deleted(self):
        return f"{self.name}_persona_deleted"

    @property
    @abstractmethod
    def model(self) -> NeonLLM:
        pass

    @create_mq_callback()
    def handle_request(self, body: dict) -> Thread:
        """
        Handles ask requests (response to prompt) from MQ to LLM
        :param body: request body (dict)
        """
        # Handle this asynchronously so multiple subminds can be handled
        # concurrently
        t = Thread(target=self._handle_request_async, args=(body,),
                   daemon=True)
        t.start()
        return t

    @create_mq_callback()
    def handle_score_request(self, body: dict):
        """
        Handles score requests (vote) from MQ to LLM
        :param body: request body (dict)
        """
        # Handle this asynchronously so multiple subminds can be handled
        # concurrently
        t = Thread(target=self._handle_score_async, args=(body,),
                   daemon=True)
        t.start()
        return t

    @create_mq_callback()
    def handle_opinion_request(self, body: dict):
        """
        Handles opinion requests (discuss) from MQ to LLM
        :param body: request body (dict)
        """
        # Handle this asynchronously so multiple subminds can be handled
        # concurrently
        t = Thread(target=self._handle_opinion_async, args=(body,),
                   daemon=True)
        t.start()
        return t

    @create_mq_callback()
    def handle_persona_update(self, body: dict):
        """
        Handles an emitted message from the server containing updated persona data
        for this LLM
        :param body: MQ message body containing persona data for update
        """
        with self._persona_update_lock:
            self._personas_provider.apply_persona_data(persona_data=body)

    @create_mq_callback()
    def handle_persona_delete(self, body: dict):
        """
        Handles an emitted message from the server containing deleted persona data
        for this LLM
        :param body: MQ message body containing persona data for deletion
        """
        with self._persona_update_lock:
            self._personas_provider.remove_persona(body)

    def _handle_request_async(self, request: dict):
        message_id = request["message_id"]
        routing_key = request["routing_key"]

        query = request["query"]
        history = request["history"]
        persona = request.get("persona", {})
        LOG.debug(f"Request persona={persona}|key={routing_key}")
        # Default response if the model fails to respond
        response = 'Sorry, but I cannot respond to your message at the '\
                   'moment; please, try again later'
        try:
            response = self.model.ask(message=query, chat_history=history,
                                      persona=persona)
        except ValueError as err:
            LOG.error(f'ValueError={err}')
        except Exception as e:
            LOG.exception(e)
        api_response = LLMProposeResponse(message_id=message_id,
                                          response=response,
                                          routing_key=routing_key)
        LOG.debug(f"Sending response: {response}")
        self.send_message(request_data=api_response.model_dump(),
                          queue=routing_key)
        LOG.info(f"Handled ask request for query={query}")

    def _handle_score_async(self, body: dict):
        """
        Handles score requests (vote) from MQ to LLM
        :param body: request body (dict)
        """
        message_id = body["message_id"]
        routing_key = body["routing_key"]

        query = body["query"]
        responses = body["responses"]
        persona = body.get("persona", {})

        if not responses:
            sorted_answer_idx = []
        else:
            try:
                sorted_answer_idx = self.model.get_sorted_answer_indexes(
                    question=query, answers=responses, persona=persona)
            except ValueError as err:
                LOG.error(f'ValueError={err}')
                sorted_answer_idx = []
            except Exception as e:
                LOG.exception(e)
                sorted_answer_idx = []

        api_response = LLMVoteResponse(message_id=message_id,
                                       routing_key=routing_key,
                                       sorted_answer_indexes=sorted_answer_idx)
        self.send_message(request_data=api_response.model_dump(),
                          queue=routing_key)
        LOG.info(f"Handled score request for query={query}")

    def _handle_opinion_async(self, body: dict):
        """
        Handles opinion requests (discuss) from MQ to LLM
        :param body: request body (dict)
        """
        message_id = body["message_id"]
        routing_key = body["routing_key"]

        query = body["query"]
        options = body["options"]
        persona = body.get("persona", {})
        responses = list(options.values())

        if not responses:
            opinion = "Sorry, but I got no options to choose from."
        else:
            # Default opinion if the model fails to respond
            opinion = "Sorry, but I experienced an issue trying to form "\
                      "an opinion on this topic"
            try:
                sorted_answer_indexes = self.model.get_sorted_answer_indexes(
                    question=query, answers=responses, persona=persona)
                best_respondent_nick, best_response = list(options.items())[
                    sorted_answer_indexes[0]]
                opinion = self._ask_model_for_opinion(
                    respondent_nick=best_respondent_nick,
                    question=query, answer=best_response, persona=persona)
            except ValueError as err:
                LOG.error(f'ValueError={err}')
            except IndexError as err:
                # Failed response will return an empty list
                LOG.error(f'IndexError={err}')
            except Exception as e:
                LOG.exception(e)

        api_response = LLMDiscussResponse(message_id=message_id,
                                          routing_key=routing_key,
                                          opinion=opinion)
        self.send_message(request_data=api_response.model_dump(),
                          queue=routing_key)
        LOG.info(f"Handled discuss request for query={query}")

    def _ask_model_for_opinion(self, respondent_nick: str, question: str,
                               answer: str, persona: dict) -> str:
        prompt = self.compose_opinion_prompt(respondent_nick=respondent_nick,
                                             question=question,
                                             answer=answer)
        opinion = self.model.ask(message=prompt, chat_history=[],
                                 persona=persona)
        LOG.info(f'Received LLM opinion={opinion}, prompt={prompt}')
        return opinion

    @staticmethod
    @abstractmethod
    def compose_opinion_prompt(respondent_nick: str, question: str,
                               answer: str) -> str:
        """
        Format a response into a prompt to evaluate another submind's response
        @param respondent_nick: Name of submind providing a response
        @param question: Prompt being responded to
        @param answer: respondent's response to the question
        """
        pass

    def run(self, run_consumers: bool = True, run_sync: bool = False,
            run_observer: Optional[bool] = None, **kwargs):
        MQConnector.run(self, run_consumers=run_consumers, run_sync=run_sync,
                        run_observer=run_observer, **kwargs)
        if not self.started:
            raise RuntimeError(f'Failed to connect to MQ. config={self.config}')
        self._personas_provider.start_sync()

    def stop(self):
        super().stop()
        self._personas_provider.stop_sync()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import List, Optional
from uuid import uuid4

from chatbot_core.v2 import ChatBot
from neon_data_models.models.api.mq import (
    LLMProposeRequest,
    LLMDiscussRequest,
    LLMVoteRequest,
    LLMProposeResponse,
    LLMDiscussResponse,
    LLMVoteResponse,
)
from neon_mq_connector.utils.client_utils import send_mq_request
from neon_utils.logger import LOG
from neon_data_models.models.api.llm import LLMPersona

from neon_llm_core.utils.config import LLMMQConfig
from neon_llm_core.utils.constants import DEFAULT_RESPONSE, DEFAULT_VOTE


class LLMBot(ChatBot):

    def __init__(self, *args, **kwargs):
        ChatBot.__init__(self, *args, **kwargs)
        self.bot_type = "submind"
        self.base_llm = kwargs["llm_name"]  # chatgpt, fastchat, etc.
        self.persona = kwargs["persona"]
        self.persona = LLMPersona(**self.persona) if \
            isinstance(self.persona, dict) else self.persona
        self.mq_queue_config = self.get_llm_mq_config(self.base_llm)
        LOG.info(f'Initialised config for llm={self.base_llm}|'
                 f'persona={self._bot_id}')
        self.prompt_id_to_shout = dict()

    @property
    def contextual_api_supported(self):
        return True

    def ask_chatbot(self, user: str, shout: str, timestamp: str,
                    context: dict = None) -> str:
        """
        Handles an incoming shout into the current conversation
        :param user: user associated with shout
        :param shout: text shouted by user
        :param timestamp: formatted timestamp of shout
        :param context: message context
        """
        prompt_id = context.get('prompt_id') if context else None
        if prompt_id:
            self.prompt_id_to_shout[prompt_id] = shout
        LOG.debug(f"Getting response to {shout}")
        response = self._get_llm_api_response(shout=shout)
        return response.response if response else DEFAULT_RESPONSE

    def ask_discusser(self, options: dict, context: dict = None) -> str:
        """
        Provides one discussion response based on the given options

        :param options: proposed responses (botname: response)
        :param context: message context
        """
        if not options:
            LOG.error(f"No options provided: options={options}")
            return DEFAULT_RESPONSE
        options = {k: v for k, v in options.items() if k != self.service_name}
        prompt_id = context.get('prompt_id') if context else None
        prompt_sentence = None
        if prompt_id:
            prompt_sentence = self.prompt_id_to_shout.get(prompt_id)
        LOG.info(f'prompt_sentence={prompt_sentence}, options={options}')
        if prompt_sentence:
            opinion = self._get_llm_api_opinion(prompt=prompt_sentence,
                                                options=options)
            return opinion.opinion if opinion else DEFAULT_RESPONSE
        return DEFAULT_RESPONSE

    def ask_appraiser(self, options: dict, context: dict = None) -> str:
        """
        Selects one of the responses to a prompt and casts a vote in the conversation.
        :param options: proposed responses (botname: response)
        :param context: message context
        """
        if not options:
            LOG.error(f"No options provided: options={options}")
            return DEFAULT_VOTE

        # Determine the relevant prompt
        prompt_id = context.get('prompt_id') if context else None
        prompt_sentence = None
        if prompt_id:
            prompt_sentence = self.prompt_id_to_shout.get(prompt_id)

        # Remove self answer from available options
        options = {k: v for k, v in options.items()
                   if k != self.service_name}

        if prompt_sentence and options:
            bots = list(options)
            bot_responses = list(options.values())
            LOG.info(f'bots={bots}, len(bot_responses)={len(bot_responses)}')
            answer_data = self._get_llm_api_choice(prompt=prompt_sentence,
                                                   responses=bot_responses)
            LOG.info(f'Received answer_data={answer_data}')
            if not answer_data:
                LOG.warning("No response to vote request")
                return DEFAULT_VOTE
            if len(answer_data.sorted_answer_indexes) != len(bots):
                LOG.error(f"Invalid vote response! "
                          f"len(bot_responses)={len(bot_responses)}|"
                          f"response_idxs={answer_data.sorted_answer_indexes}")
                return DEFAULT_VOTE
            if answer_data and answer_data.sorted_answer_indexes:
                return bots[answer_data.sorted_answer_indexes[0]]
        return DEFAULT_VOTE

    def _get_llm_api_response(self, shout: str) -> Optional[LLMProposeResponse]:
        """
        Requests LLM API for response on provided shout
        :param shout: Input prompt to respond to
        :returns response from LLM API
        """
        queue = self.mq_queue_config.ask_response_queue
        response_queue = f"{queue}.response.{uuid4().hex}"
        try:
            LOG.info(f"Sending to {self.mq_queue_config.vhost}/{queue} for "
                     f"persona={self.persona.name}|response_q={response_queue}")

            request_data = LLMProposeRequest(model=self.base_llm,
                                             persona=self.persona,
                                             query=shout,
                                             history=[],
                                             message_id="")
            resp_data = send_mq_request(vhost=self.mq_queue_config.vhost,
                                        request_data=request_data.model_dump(),
                                        target_queue=queue,
                                        response_queue=response_queue)
            if not resp_data:
                LOG.warning(f"Timed out waiting for response on "
                            f"{response_queue}")
                return None
            LOG.info(f"Got response for persona={self.persona}")
            return LLMProposeResponse.model_validate(obj=resp_data)
        except Exception as e:
            LOG.exception(f"Failed to get response on "
                          f"{self.mq_queue_config.vhost}/{queue}: {e}")

    def _get_llm_api_opinion(self, prompt: str, options: dict) -> Optional[LLMDiscussResponse]:
        """
        Requests LLM API for discussion of provided submind responses
        :param prompt: incoming prompt text
        :param options: proposed responses (botname: response)
        :returns response data from LLM API
        """
        queue = self.mq_queue_config.ask_discusser_queue
        response_queue = f"{queue}.response.{uuid4().hex}"

        try:
            LOG.info(f"Sending to {self.mq_queue_config.vhost}/{queue} for "
                     f"persona={self.persona.name}|response_q={response_queue}")

            request_data = LLMDiscussRequest(model=self.base_llm,
                                             persona=self.persona,
                                             query=prompt,
                                             options=options,
                                             history=[],
                                             message_id="")
            resp_data = send_mq_request(vhost=self.mq_queue_config.vhost,
                                        request_data=request_data.model_dump(),
                                        target_queue=queue,
                                        response_queue=response_queue)
            if not resp_data:
                LOG.warning(f"Timed out waiting for response on "
                            f"{response_queue}")
                return None
            return LLMDiscussResponse.model_validate(obj=resp_data)
        except Exception as e:
            LOG.exception(f"Error getting response on "
                          f"{self.mq_queue_config.vhost}/{queue}: {e}")

    def _get_llm_api_choice(self, prompt: str,
                            responses: List[str]) -> Optional[LLMVoteResponse]:
        """
        Requests LLM API for choice among provided message list
        :param prompt: incoming prompt text
        :param responses: list of answers to select from
        :returns response data from LLM API
        """
        queue = self.mq_queue_config.ask_appraiser_queue
        response_queue = f"{queue}.response.{uuid4().hex}"

        try:
            LOG.info(f"Sending to {self.mq_queue_config.vhost}/{queue} for "
                     f"persona={self.persona.name}|response_q={response_queue}")

            request_data = LLMVoteRequest(model=self.base_llm,
                                          persona=self.persona,
                                          query=prompt,
                                          responses=responses,
                                          history=[],
                                          message_id="")
            resp_data = send_mq_request(vhost=self.mq_queue_config.vhost,
                                        request_data=request_data.model_dump(),
                                        target_queue=queue,
                                        response_queue=response_queue)
            if not resp_data:
                LOG.warning(f"Timed out waiting for response on "
                            f"{response_queue}")
                return None
            return LLMVoteResponse.model_validate(obj=resp_data)
        except Exception as e:
            LOG.exception(f"Failed to get response on "
                          f"{self.mq_queue_config.vhost}/{queue}: {e}")

    @staticmethod
    def get_llm_mq_config(llm_name: str) -> LLMMQConfig:
        """
        Get MQ queue names that the LLM service has access to. These are
        LLM-oriented, not bot/persona-oriented.
        """
        return LLMMQConfig(ask_response_queue=f"{llm_name}_input",
                           ask_appraiser_queue=f"{llm_name}_score_input",
                           ask_discusser_queue=f"{llm_name}_discussion_input")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 NeonGecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from unittest import TestCase
from unittest.mock import patch

from neon_data_models.models.api import (
    LLMPersona,
    LLMProposeRequest,
    LLMProposeResponse,
    LLMDiscussRequest,
    LLMDiscussResponse,
    LLMVoteRequest,
    LLMVoteResponse,
)

from neon_llm_core.chatbot import LLMBot
from neon_llm_core.utils.config import LLMMQConfig


class MockChatbot(LLMBot):
    def __init__(self):
        LLMBot.__init__(self, llm_name="mock_chatbot",
                        persona={"name": "test_persona",
                                 "system_prompt": "Test Prompt"})


class TestChatbot(TestCase):
    mock_chatbot = MockChatbot()

    @classmethod
    def tearDownClass(cls):
        cls.mock_chatbot.shutdown()

    def test_00_init(self):
        self.assertEqual(self.mock_chatbot.bot_type, "submind")
        self.assertIsInstance(self.mock_chatbot.base_llm, str)
        self.assertIsInstance(self.mock_chatbot.persona, LLMPersona)
        self.assertIsInstance(self.mock_chatbot.mq_queue_config, LLMMQConfig)

    @patch.object(mock_chatbot, '_get_llm_api_response')
    def test_ask_chatbot(self, get_api_response):
        get_api_response.return_value = LLMProposeResponse(message_id="",
                                                           response="test_resp")
        valid_prompt_id = "test_prompt_id_ask"
        valid_user = "test_user"
        valid_shout = "test_shout"
        valid_timestamp = datetime.now().isoformat()
        valid_context = {"prompt_id": valid_prompt_id}

        # Valid Request
        resp = self.mock_chatbot.ask_chatbot(valid_user, valid_shout,
                                             valid_timestamp, valid_context)
        get_api_response.assert_called_with(shout=valid_shout)
        self.assertEqual(resp, "test_resp")
        self.assertEqual(self.mock_chatbot.prompt_id_to_shout[valid_prompt_id],
                         valid_shout)

        # Valid without context
        resp = self.mock_chatbot.ask_chatbot(valid_user, valid_shout,
                                             valid_timestamp)
        get_api_response.assert_called_with(shout=valid_shout)
        self.assertEqual(resp, "test_resp")
        self.assertEqual(self.mock_chatbot.prompt_id_to_shout[valid_prompt_id],
                         valid_shout)

        # Invalid request
        self.assertIsInstance(self.mock_chatbot.ask_chatbot(valid_user,
                                                            None,
                                                            valid_timestamp),
                              str)
        get_api_response.assert_called_with(shout=None)

        # Invalid response
        get_api_response.return_value = None
        from neon_llm_core.utils.constants import DEFAULT_RESPONSE
        self.assertEqual(self.mock_chatbot.ask_chatbot(valid_user,
                                                       valid_shout,
                                                       valid_timestamp,
                                                       valid_context),
                         DEFAULT_RESPONSE)
        get_api_response.assert_called_with(shout=valid_shout)

    @patch.object(mock_chatbot, '_get_llm_api_opinion')
    def test_ask_discusser(self, get_api_opinion):
        get_api_opinion.return_value = LLMDiscussResponse(message_id="",
                                                          opinion="test_resp")
        valid_prompt_id = "test_prompt_id_disc"
        valid_prompt = "test prompt"
        valid_options = {"bot 1": "response 1", "bot 2": "response 2"}
        valid_context = {"prompt_id": valid_prompt_id}

        self.mock_chatbot.prompt_id_to_shout[valid_prompt_id] = valid_prompt

        # Valid request
        resp = self.mock_chatbot.ask_discusser(valid_options, valid_context)
        get_api_opinion.assert_called_with(prompt=valid_prompt,
                                           options=valid_options)
        self.assertEqual(resp, "test_resp")

        # Invalid response
        get_api_opinion.return_value = None
        from neon_llm_core.utils.constants import DEFAULT_RESPONSE
        self.assertEqual(self.mock_chatbot.ask_discusser(valid_options,
                                                         valid_context),
                         DEFAULT_RESPONSE)
        get_api_opinion.assert_called_with(prompt=valid_prompt,
                                           options=valid_options)

    @patch.object(mock_chatbot, '_get_llm_api_choice')
    def test_ask_appraiser(self, get_api_choice):
        get_api_choice.return_value = LLMVoteResponse(
            message_id="", sorted_answer_indexes=[2, 0, 1])
        valid_prompt_id = "test_prompt_id_vote"
        valid_prompt = "test prompt"
        options = {"bot 0": "response 0",
                   "bot 1": "response 1",
                   "bot 2": "response 2",
                   self.mock_chatbot.service_name: "Self response"}
        valid_options = ["response 0", "response 1", "response 2"]
        valid_context = {"prompt_id": valid_prompt_id}

        self.mock_chatbot.prompt_id_to_shout[valid_prompt_id] = valid_prompt

        # Valid request
        resp = self.mock_chatbot.ask_appraiser(options, valid_context)
        get_api_choice.assert_called_with(prompt=valid_prompt,
                                          responses=valid_options)
        self.assertEqual(resp, "bot 2")

        # Invalid no valid options
        resp = self.mock_chatbot.ask_appraiser(
            {self.mock_chatbot.service_name: "Self response"},
            valid_context)
        self.assertIn("abstain", resp.lower())

        # Invalid API response
        get_api_choice.reset_mock()
        get_api_choice.return_value = None
        resp = self.mock_chatbot.ask_appraiser(options, valid_context)
        get_api_choice.assert_called_with(prompt=valid_prompt,
                                          responses=valid_options)
        self.assertIn("abstain", resp.lower())

    @patch('neon_llm_core.chatbot.send_mq_request')
    def test_get_llm_api_response(self, mq_request):
        mq_request.return_value = {"response": "test",
                                   "message_id": ""}

        # Valid Request
        resp = self.mock_chatbot._get_llm_api_response("input")
        request_data = mq_request.call_args.kwargs['request_data']
        req = LLMProposeRequest(**request_data)
        self.assertIsInstance(req, LLMProposeRequest)
        self.assertEqual(req.query, "input")
        self.assertEqual(req.model, self.mock_chatbot.base_llm)
        self.assertEqual(req.persona, self.mock_chatbot.persona)
        self.assertIsInstance(resp, LLMProposeResponse)
        self.assertEqual(resp.response, mq_request.return_value['response'])

        # Invalid request
        self.assertIsNone(self.mock_chatbot._get_llm_api_response(None))

        # Invalid response
        mq_request.return_value = {}
        self.assertIsNone(self.mock_chatbot._get_llm_api_response("input"))

    @patch('neon_llm_core.chatbot.send_mq_request')
    def test_get_llm_api_opinion(self, mq_request):
        mq_request.return_value = {"opinion": "test",
                                   "message_id": ""}
        prompt = "test prompt"
        options = {"bot 1": "resp 1", "bot 2": "resp 2"}

        # Valid Request
        resp = self.mock_chatbot._get_llm_api_opinion(prompt, options)
        request_data = mq_request.call_args.kwargs['request_data']
        req = LLMDiscussRequest(**request_data)
        self.assertIsInstance(req, LLMDiscussRequest)
        self.assertEqual(req.query, prompt)
        self.assertEqual(req.options, options)
        self.assertEqual(req.model, self.mock_chatbot.base_llm)
        self.assertEqual(req.persona, self.mock_chatbot.persona)
        self.assertIsInstance(resp, LLMDiscussResponse)
        self.assertEqual(resp.opinion, mq_request.return_value['opinion'])

        # Invalid request
        self.assertIsNone(self.mock_chatbot._get_llm_api_opinion(prompt,
                                                                 prompt))

        # Invalid response
        mq_request.return_value = {}
        self.assertIsNone(self.mock_chatbot._get_llm_api_opinion(prompt,
                                                                 options))

    @patch('neon_llm_core.chatbot.send_mq_request')
    def test_get_llm_api_choice(self, mq_request):
        mq_request.return_value = {"sorted_answer_indexes": [2, 0, 1],
                                   "message_id": ""}
        prompt = "test prompt"
        responses = ["one", "two", "three"]

        # Valid Request
        resp = self.mock_chatbot._get_llm_api_choice(prompt, responses)
        request_data = mq_request.call_args.kwargs['request_data']

        req = LLMVoteRequest(**request_data)
        self.assertIsInstance(req, LLMVoteRequest)
        self.assertEqual(req.query, prompt)
        self.assertEqual(req.responses, responses)
        self.assertEqual(req.model, self.mock_chatbot.base_llm)
        self.assertEqual(req.persona, self.mock_chatbot.persona)
        self.assertIsInstance(resp, LLMVoteResponse)
        self.assertEqual(resp.sorted_answer_indexes,
                         mq_request.return_value['sorted_answer_indexes'])

        # Invalid request
        self.assertIsNone(self.mock_chatbot._get_llm_api_choice(prompt,
                                                                [1, 2, 3]))

        # Invalid response
        mq_request.return_value["sorted_answer_indexes"] = ["one", "two",
                                                            "three"]
        self.assertIsNone(self.mock_chatbot._get_llm_api_choice(prompt,
                                                                responses))



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 NeonGecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from unittest.mock import patch, Mock, MagicMock

from neon_llm_core.utils.personas.provider import PersonasProvider
from neon_llm_core.utils.personas.state import PersonaHandlersState

from .utils.factory import PersonaFactory


class TestPersonasProvider(unittest.TestCase):

    def setUp(self):
        """Set up the test environment and mock dependencies."""
        self.mock_service_name = "mock_service"
        self.mock_config = {
            "llm_bots": {
                self.mock_service_name: [PersonaFactory.create_mock_llm_persona(enabled=True) for _ in range(5)]
            },
            "MQ": {"users": {"neon_llm_submind": {"user": "test", "password": "test"}}}
        }
        self.provider = PersonasProvider(service_name=self.mock_service_name, ovos_config=self.mock_config)

    @patch("neon_llm_core.utils.personas.provider.send_mq_request")
    @patch("neon_llm_core.utils.personas.state.LLMBot")
    def test_fetch_persona_config_success(self, mock_bot, mock_send_mq_request):
        mock_persona_data = [
            PersonaFactory.create_mock_llm_persona(enabled=True).model_dump(),
            PersonaFactory.create_mock_llm_persona(enabled=False).model_dump(),
        ]
        mock_send_mq_request.return_value = {"items": mock_persona_data}
        self.provider._fetch_persona_config()

        self.assertEqual(len(self.provider.personas), 1)
        self.assertEqual(self.provider.personas[0].name, mock_persona_data[0]["name"])

    @patch("neon_llm_core.utils.personas.provider.send_mq_request")
    @patch("neon_llm_core.utils.personas.state.LLMBot")
    def test_fetch_persona_config_empty(self, mock_bot, mock_send_mq_request):
        mock_send_mq_request.return_value = {"items": []}
        self.provider._fetch_persona_config()

        self.assertEqual(len(self.provider.personas), 0)

    def test__validate_persona_data_success(self):
        persona_data = PersonaFactory.create_mock_llm_persona().model_dump()
        persona = self.provider._validate_persona_data(persona_data)

        self.assertIsNotNone(persona)
        self.assertEqual(persona.name, persona_data['name'])

    def test__validate_persona_data_returns_None_on_failure(self):
        persona_data = {}
        persona = self.provider._validate_persona_data(persona_data)

        self.assertIsNone(persona)

    def test__validate_persona_identity_success(self):
        persona_identity = PersonaFactory.create_mock_persona_identity().model_dump()
        persona = self.provider._validate_persona_identity(persona_identity)

        self.assertIsNotNone(persona)
        self.assertEqual(persona.name, persona_identity['name'])

    def test__validate_persona_identity_returns_None_on_failure(self):
        persona_identity_data = {}
        persona_identity = self.provider._validate_persona_identity(persona_identity_data)

        self.assertIsNone(persona_identity)

    @patch.object(PersonaHandlersState, "add_persona_handler")
    def test__add_persona_success(self, mock_add_persona_handler):
        mock_add_persona_handler.return_value = True
        persona = Mock(id="mock_persona")

        result = self.provider._add_persona(persona)

        self.assertTrue(result)
        mock_add_persona_handler.assert_called_once_with(persona=persona)

    @patch("neon_llm_core.utils.personas.state.LLMBot")
    def test__add_persona_when_default_personas_are_running(self, mock_llm_bot):
        persona = PersonaFactory.create_mock_llm_persona(enabled=True)

        self.provider._persona_handlers_state.init_default_personas()
        self.assertEqual(self.provider._persona_handlers_state.connected_persona_ids,
                         [persona.id for persona in self.mock_config['llm_bots'][self.mock_service_name]])

        result = self.provider._add_persona(persona)
        self.assertTrue(result)

        # Defaults personas eliminated
        self.assertEqual(self.provider._persona_handlers_state.connected_persona_ids,
                         [persona.id])

    @patch("neon_llm_core.utils.personas.state.LLMBot")
    def test__add_disabled_persona_when_default_personas_are_running(self, mock_llm_bot):
        persona = PersonaFactory.create_mock_llm_persona(enabled=False)

        self.provider._persona_handlers_state.init_default_personas()
        self.assertEqual(self.provider._persona_handlers_state.connected_persona_ids,
                         [persona.id for persona in self.mock_config['llm_bots'][self.mock_service_name]])

        result = self.provider._add_persona(persona)
        self.assertFalse(result)

        # Defaults personas stay as is
        self.assertEqual(self.provider._persona_handlers_state.connected_persona_ids,
                         [persona.id for persona in self.mock_config['llm_bots'][self.mock_service_name]])

    @patch("neon_llm_core.utils.personas.state.LLMBot")
    def test_disable_latest_persona_should_activate_default_personas(self, mock_llm_bot):
        persona = PersonaFactory.create_mock_llm_persona(enabled=True)

        self.provider._add_persona(persona)

        # Persona is connected
        self.assertEqual(self.provider._persona_handlers_state.connected_persona_ids,
                         [persona.id])

        persona.enabled = False

        self.provider._add_persona(persona)

        # Persona is disconnected and default personas are initialised
        self.assertEqual(self.provider._persona_handlers_state.connected_persona_ids,
                         [persona.id for persona in self.mock_config['llm_bots'][self.mock_service_name]])

    def test_should_reset_personas_when_no_data(self):
        result = self.provider._should_reset_personas([])
        self.assertTrue(result)

    def test_should_not_reset_personas_with_data_and_no_sync_interval(self):
        self.provider.PERSONA_SYNC_INTERVAL = 0
        result = self.provider._should_reset_personas([MagicMock()])
        self.assertFalse(result)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 NeonGecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from unittest.mock import MagicMock, patch

from neon_llm_core.utils.personas.state import PersonaHandlersState
from .utils.factory import PersonaFactory


class TestPersonaHandlersState(unittest.TestCase):

    def setUp(self):
        self.mock_service_name = "test_service"
        self.mock_config = {
            "llm_bots": {
                self.mock_service_name: [PersonaFactory.create_mock_llm_persona(enabled=True) for _ in range(5)]
            },
            "MQ": {"users": {"neon_llm_submind": {"user": "test", "password": "test"}}}
        }
        self.persona_handlers_state = PersonaHandlersState(
            service_name=self.mock_service_name,
            ovos_config=self.mock_config
        )

    def test_initialization(self):
        self.assertEqual(self.persona_handlers_state.service_name, self.mock_service_name)
        self.assertEqual(self.persona_handlers_state.ovos_config, self.mock_config)
        self.assertFalse(self.persona_handlers_state.default_personas_running)
        self.assertDictEqual(self.persona_handlers_state._created_items, {})

    def test_default_personas(self):
        self.assertEqual(
            self.persona_handlers_state.default_personas,
            self.mock_config["llm_bots"][self.mock_service_name]
        )

    def test_has_connected_personas(self):
        self.assertFalse(self.persona_handlers_state.has_connected_personas())
        self.persona_handlers_state._created_items["persona_1"] = MagicMock()
        self.assertTrue(self.persona_handlers_state.has_connected_personas())

    @patch("neon_llm_core.utils.personas.state.LLMBot")
    def test_init_default_personas(self, mock_bot):
        self.persona_handlers_state.init_default_personas()
        self.assertTrue(self.persona_handlers_state.default_personas_running)

    @patch("neon_llm_core.utils.personas.state.LLMBot")
    @patch("neon_data_models.models.LLMPersona", new_callable=PersonaFactory.create_mock_llm_persona, enabled=True)
    def test_add_persona_handler(self, mock_persona, mock_bot):
        result = self.persona_handlers_state.add_persona_handler(mock_persona)
        self.assertIn(mock_persona.id, self.persona_handlers_state._created_items)
        self.assertEqual(result, mock_bot())

    def test_clean_up_all_personas(self):
        mock_personas = [PersonaFactory.create_mock_llm_persona(enabled=True) for _ in range(5)]
        self.persona_handlers_state._created_items = {mock_persona.id: MagicMock()
                                                      for mock_persona in mock_personas}
        self.persona_handlers_state.clean_up_personas()

        self.assertEqual(len(self.persona_handlers_state._created_items), 0)

    def test_clean_up_personas_keeps_specified_item(self):
        mock_personas = [PersonaFactory.create_mock_llm_persona(enabled=True) for _ in range(5)]
        self.persona_handlers_state._created_items = {mock_persona.id: MagicMock()
                                                      for mock_persona in mock_personas}

        persona_to_keep = mock_personas[0]
        self.persona_handlers_state.clean_up_personas(ignore_items=[persona_to_keep])

        self.assertEqual(len(self.persona_handlers_state._created_items), 1)
        self.assertTrue(persona_to_keep.id in self.persona_handlers_state._created_items)

    def test_remove_persona(self):
        mock_item = MagicMock()
        self.persona_handlers_state._created_items = {"persona_1": mock_item}
        self.persona_handlers_state.remove_persona("persona_1")
        mock_item.stop.assert_called_once()
        self.assertNotIn("persona_1", self.persona_handlers_state._created_items)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from functools import cached_property
from threading import Lock
from typing import Dict, List, Optional

from neon_data_models.models.api.llm import LLMPersona
from neon_utils.logger import LOG
from neon_llm_core.chatbot import LLMBot


class PersonaHandlersState:
    """
    This works with the PersonasProvider object to manage LLMBot instances for
    all configured personas.
    """

    def __init__(self, service_name: str, ovos_config: dict):
        self._created_items: Dict[str, LLMBot] = {}
        self.service_name = service_name
        self.ovos_config = ovos_config
        self.mq_config = ovos_config.get('MQ', {})
        self.default_personas_running = False

        self.personas_clean_up_lock = Lock()
        self.personas_remove_lock = Lock()

    @cached_property
    def default_personas(self):
        return self.ovos_config.get("llm_bots", {}).get(self.service_name, [])

    @property
    def connected_persona_ids(self) -> List[str]:
        return list(self._created_items)

    def has_connected_personas(self) -> bool:
        return bool(self._created_items)

    def init_default_personas(self):
        """
        Initializes LLMBot instances for all personas defined in configuration.
        """
        if self.default_personas and not self.default_personas_running:
            self.clean_up_personas()
            LOG.info(f"Initializing default personas for: {self.service_name}")
            for persona in self.default_personas:
                self.add_persona_handler(
                    persona=LLMPersona.model_validate(obj=persona)
                )
            self.default_personas_running = True
        else:
            if self.default_personas_running:
                LOG.debug('Default personas already running')
            elif not self.default_personas:
                LOG.warning('Default personas not configured')

    def add_persona_handler(self, persona: LLMPersona) -> Optional[LLMBot]:
        """
        Creates an `LLMBot` instance for the given persona if the persona does
        not yet exist AND the persona is not disabled in configuration.
        :param persona: Persona definition to generate an LLMBot instance of
        :return: New or existing LLMBot instance or None if the persona is
                 disabled
        """
        persona_dict = persona.model_dump()
        if persona.id in list(self._created_items):
            if self._created_items[persona.id].persona != persona_dict:
                LOG.info(f"Received new data for persona: '{persona.id}' - removing old instance")
                self.remove_persona(persona_id=persona.id)
            else:
                LOG.debug('Persona config provided is identical to existing, skipping')
                return
        if not persona.enabled:
            LOG.warning(f"Skipping disabled persona: '{persona.id}'")
            return
        # Get a configured username to use for LLM submind connections
        persona_id = f"{persona.id}_{self.service_name}"
        self.ovos_config["MQ"]["users"][persona_id] = self.mq_config['users']['neon_llm_submind']
        bot = LLMBot(llm_name=self.service_name, service_name=persona_id,
                     persona=persona_dict, config=self.ovos_config,
                     vhost="/chatbots")
        bot.run()
        LOG.info(f"Started chatbot: {bot.service_name}")
        self._created_items[persona.id] = bot
        return bot

    def clean_up_personas(self, ignore_items: List[LLMPersona] = None):
        with self.personas_clean_up_lock:
            connected_personas = set(self._created_items)
            ignored_persona_ids = set(persona.id for persona in ignore_items or [])
            personas_to_remove = connected_personas - ignored_persona_ids
            for persona_id in personas_to_remove:
                self.remove_persona(persona_id=persona_id)

    def remove_persona(self, persona_id: str):
        with self.personas_remove_lock:
            if persona_id in self._created_items:
                try:
                    LOG.info(f'Removing persona_id = {persona_id}')
                    self._created_items[persona_id].stop()
                    self._created_items.pop(persona_id, None)
                except Exception as ex:
                    LOG.warning(f'Failed to gracefully stop persona={persona_id!r}, ex={str(ex)}')
